// Board Library : Seeed nRF52 mbed-enable Borads 2.8.1
// Board Select  : Seeed nRF52 mbed-enable Borads / Seeed XIAO BLE - nRF52840
// 2022/12/23      Sometimes Peripheral disconnections may not be detected
//                 Sometimes RSSI values are not being retrieved properly
//                 Sometimes connection takes a long time --> need to reset central and peripherals
//**********************************************************************************************
// [IMPORTANT] Battery voltage not readable bug fix
// #define analogPinToPinName(P)  
// /Arduino15/packages/Seeeduino/hardware/mbed/2.7.2/cores/arduino/pinDefinitions.h:22
// g_APinDescription ---> g_AAnalogPinDescription
//**********************************************************************************************
// peripheral : nRF52_XIAO_Attitude_peripheral_1.ino
//              nRF52_XIAO_Attitude_peripheral_2.ino

#include <Wire.h>
#include <Adafruit_GFX.h>         // Adafruit GFX Library 1.10.14 by Adafruit
#include "Fonts/FreeSans9pt7b.h"
#include "Fonts/FreeSerif9pt7b.h"
#include <Adafruit_SSD1306.h>     // Adafruit SSD1306 2.5.3 by Adafruit
#include <ArduinoBLE.h>           // ArduinoBLE 1.3.1 by Arduino

#define VBAT_LOWER  3.5           // battery voltage lower limit
#define VBAT_UPPER  4.2           // battery voltage upper limit
#define dataNum 8                 // receive data number : roll, pitch, yaw, Vbatt (4 data 8 byte)

Adafruit_SSD1306 display(128, 64, &Wire, -1);   //SSD1306

const char* versionNumber = "0.90"; // version number
int err;                            // error code of scan_connect function
bool LED_state;                     // LED ON/OFF state 

struct prphData {                   // data struct for peripheral
  float roll, pitch, yaw;             // attitude
  float Vbatt;                        // battery voltage
  int ss;                             // peripheral signal strength
  bool readingFlag = false;           // data buffer in use flag
  union unionData {                   // Union for bit convertion 16 <--> 8
    int16_t   dataBuff16[dataNum/2];
    uint8_t   dataBuff8[dataNum];
  };
  union unionData ud;
};
prphData sd_1;                      // data struct for peripheral 1
prphData sd_2;                      // data struct for peripheral 2

// UUID and Device
BLEDevice peripheral_1;
#define myUUID_1(val) ("0dd7eb5a-" val "-4f45-bcd7-94c674c3b25f")
//BLEService        AttService_1(myUUID_1("0000"));
BLECharacteristic dataCharacteristic_1(myUUID_1("0010"), BLEWritten, dataNum);

BLEDevice peripheral_2;
#define myUUID_2(val) ("f78d0557-" val "-4cbc-9ab3-89fc6514877f")
//BLEService        AttService_2(myUUID_2("0000"));
BLECharacteristic dataCharacteristic_2(myUUID_2("0010"), BLEWritten, dataNum);

//***************************************************************************************
void setup() {
  //initialize serial port
  Serial.begin(115200);
  while (!Serial) { delay(1); }

  //set I/O pins
  pinMode(LED_RED, OUTPUT);       // LOW:LED ON, HIGH:LED OFF
  pinMode(LED_GREEN, OUTPUT);
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);

  // initialize SSD1306
  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {
    Serial.println("SSD1306 initialize error");
  }
  display.clearDisplay();
  display.setFont(&FreeSerif9pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("ATT Monitor");
  display.setCursor(0, 31);
  display.println("     BLE Central");
  display.setCursor(0, 47);
  display.print("Ver   : ");
  display.println(versionNumber);
  display.display();
  delay(2000);

  // initialize BLE
  BLE.begin(); 
}

//******************************************************************************************
// IMU data is read every 80mS
//
void loop() {
  // scan and connect the peripheral
  // return value of this function is an error code  
  do {   
    err = scan_connect();
    Serial.print("ERROR : ");
    Serial.println(err);
  } while(err != 0);

    // main task while connected to the peripheral 60~78mS
    while (peripheral_1.connected() & peripheral_2.connected()) {
      LED_state = !LED_state;
      digitalWrite(LED_GREEN, (LED_state ? LOW : HIGH));   // connect indicator blinking
      
      long timestamp = millis();    // for loop timer
      
      // restoration of received data
      // prohibit event handler access while data buffer is in use          
      sd_1.readingFlag = true;   
        sd_1.roll = sd_1.ud.dataBuff16[0] * 180.0 / 32768.0;
        sd_1.pitch = sd_1.ud.dataBuff16[1] * 180.0 / 32768.0;
        sd_1.yaw = sd_1.ud.dataBuff16[2] * 360.0 / 32768.0;
        sd_1.Vbatt = sd_1.ud.dataBuff16[3] / 1000.0;
      sd_1.readingFlag = false;
  
      sd_2.readingFlag = true;   
        sd_2.roll = sd_2.ud.dataBuff16[0] * 180.0 / 32768.0;
        sd_2.pitch = sd_2.ud.dataBuff16[1] * 180.0 / 32768.0;
        sd_2.yaw = sd_2.ud.dataBuff16[2] * 360.0 / 32768.0;
        sd_2.Vbatt = sd_2.ud.dataBuff16[3] / 1000.0;
      sd_2.readingFlag = false;
      
      // for serial plotter
      Serial.print(sd_1.roll); Serial.print(", ");
      Serial.print(sd_1.pitch); Serial.print(", ");
      Serial.print(sd_1.yaw); Serial.print(", ");     
      Serial.print(sd_2.roll); Serial.print(", ");
      Serial.print(sd_2.pitch); Serial.print(", ");
      Serial.print(sd_2.yaw); Serial.print("    ");
//      Serial.print(sd_1.Vbatt); Serial.print(", ");
//      Serial.print(sd_2.Vbatt); Serial.print("    ");
//      Serial.print(sd_1.ss); Serial.print(", ");
//      Serial.print(sd_2.ss);            
      Serial.println("");
   
      // display Roll, Pitch, battery voltage, signal strength
      display.clearDisplay();      
      display.setCursor(0, 15); display.print(sd_1.roll);
      display.setCursor(64, 15); display.println(sd_2.roll);
      display.setCursor(0, 31); display.print(sd_1.pitch);
      display.setCursor(64, 31); display.println(sd_2.pitch);
      
      display.setCursor(0, 47); display.print(sd_1.Vbatt);
      if(sd_1.Vbatt < VBAT_LOWER) {
        display.print("L");
      }      
      display.setCursor(64, 47); display.println(sd_2.Vbatt);
      if(sd_2.Vbatt < VBAT_LOWER) {
        display.print("L");
      }
               
      display.setCursor(0, 63); display.print(sd_1.ss);
      display.setCursor(64, 63); display.println(sd_2.ss);     
      display.display();

      while(millis() - timestamp < 80);     // wait for loop time 80mS
    } //While connected


  //if disconnected, scan_connect again
  Serial.print(peripheral_1.connected());
  Serial.print(peripheral_2.connected());
  Serial.println("  disconnect, and scan_connect again");
  BLE.disconnect();

  digitalWrite(LED_GREEN, HIGH);            // connect indicator OFF
      
} //loop

//**************************************************************************************************
// scan and connect peripherals
// return value of this function is an error code
//
int scan_connect(void) {

  // scanning peripheral_1   +++++++++++++++++++++++++++++++++++++++++++++++++
  BLE.scanForUuid(myUUID_1("0000"));
  Serial.println("1_1.SCANNING ................");
 
  peripheral_1 = BLE.available();
  if (!peripheral_1) {
    Serial.println("1_2x.Peripheral unavailable");
    return 12;
  }
  Serial.println("1_2.Peripheral is available");

  if (peripheral_1.localName() != "Att_Monitor_1") {
    Serial.println("1_3x.Peripheral local name miss match");
    return 13;
  }
  Serial.println("1_3.Got the right peripheral");
  
  // scanning peripheral_2   +++++++++++++++++++++++++++++++++++++++++++++++++++++ 
  BLE.scanForUuid(myUUID_2("0000"));
  Serial.println("2_1.SCANNING ................");
  
  peripheral_2 = BLE.available();
  if (!peripheral_2) {
    Serial.println("2_2x.Peripheral unavailable");
    return 22;
  }
  Serial.println("2_2.Peripheral is available");

  if (peripheral_2.localName() != "Att_Monitor_2") {
    Serial.println("2_3x.Peripheral local name miss match");
    return 23;
  }
  Serial.println("2_3.Got the right peripheral");

  // stop scanning, connect the peripheral  ++++++++++++++++++++++++++++++++++++++++++++++
  BLE.stopScan();
  Serial.println("4.Stop scanning");
  
  // connect the peripheral_1   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Serial.println("1_5.CONNECTING ................");
  if (!peripheral_1.connect()) {
    Serial.println("1_5x.Can't connect");
    return 15;
  } 
  Serial.println("1_5.Connected");

  if (!peripheral_1.discoverAttributes()) {
    Serial.println("1_6x.Didn't discover attributes");
    peripheral_1.disconnect();
    return 16;
  }
  Serial.println("1_6.Discovered attributes");

  dataCharacteristic_1 = peripheral_1.characteristic(myUUID_1("0010"));       //dataCaracteristic UUID  
  dataCharacteristic_1.setEventHandler(BLEWritten, characteristicWritten_1);  //BLEWritten handler
  Serial.println("1_7.Char and eventhandler setting");  

  if (!dataCharacteristic_1.canSubscribe()) {
    Serial.println("1_8x.Can't subscribe");
    peripheral_1.disconnect();
    return 18;
  }
  Serial.println("1_8.Can subscribe");

  if (!dataCharacteristic_1.subscribe()) {
    Serial.println("1_9x.Can't Subscribe");
    peripheral_1.disconnect();
    return 19;
  }
  Serial.println("1_9.Subscribed");

  // connect the peripheral_2   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Serial.println("2_5.CONNECTING ................");
  if (!peripheral_2.connect()) {
    Serial.println("2_5x.Can't connect");
    return 25;
  } 
  Serial.println("2_5.Connected");
  
  if (!peripheral_2.discoverAttributes()) {
    Serial.println("2_6x.Didn't discover attributes");
    peripheral_2.disconnect();
    return 26;
  }
  Serial.println("2_6.Discovered attributes");

  dataCharacteristic_2 = peripheral_2.characteristic(myUUID_2("0010"));       //dataCaracteristic UUID
  dataCharacteristic_2.setEventHandler(BLEWritten, characteristicWritten_2);  //BLEWritten handler
  Serial.println("2_7.Char and eventhandler setting");  

  if (!dataCharacteristic_2.canSubscribe()) {
    Serial.println("2_8x.Can't subscribe");
    peripheral_2.disconnect();
    return 28;
  }
  Serial.println("2_8.Can subscribe");

  if (!dataCharacteristic_2.subscribe()) {
    Serial.println("1_9x.Can't Subscribe");
    peripheral_2.disconnect();
    return 29;
  }
  Serial.println("2_9.Subscribed");

  Serial.println("10.Success scanning and connecting");
  return 0;
}

//****************************************************************************************************
// Characteristic written event handler for peripheral 1
//
void characteristicWritten_1(BLEDevice peripheral_1, BLECharacteristic thisChar) {
  digitalWrite(LED_RED, LOW);         // event indicator ON
    
  // wait while data buffer is accessed in main loop  
  while(sd_1.readingFlag == true) {
  }    
  dataCharacteristic_1.readValue(sd_1.ud.dataBuff8, dataNum);  // read data packet  
  int ss = peripheral_1.rssi();                                // exclude meaningless 0 and 127
  if ((ss != 0) & (ss != 127)) {
    sd_1.ss = ss;
  }  

  long timestamp = millis();
  while(millis() - timestamp <= 1);   // Delay to make LED visible
    
  digitalWrite(LED_RED, HIGH);        // event indicator OFF
}

//****************************************************************************************************
// Characteristic written event handler for peripheral 2
//
void characteristicWritten_2(BLEDevice peripheral_2, BLECharacteristic thisChar) {
  digitalWrite(LED_RED, LOW);         // event indicator ON
    
  // wait while data buffer is accessed in main loop  
  while(sd_2.readingFlag == true) {
  }    
  dataCharacteristic_2.readValue(sd_2.ud.dataBuff8, dataNum);  // read data packet  
  int ss = peripheral_2.rssi();
  if ((ss != 0) & (ss != 127)) {                               // exclude meaningless 0 and 127
    sd_2.ss = ss;
  }

  long timestamp = millis();
  while(millis() - timestamp <= 1);   // Delay to make LED visible
    
  digitalWrite(LED_RED, HIGH);        // event indicator OFF
}
